__all__ = ["nms_numba"]

import numba
import numpy as np


@numba.jit(nopython=True)
def nms_numba(boxes: np.ndarray, scores: np.ndarray, threshold: float) -> np.ndarray:
    """Non-maximum suppression with numba (faster when #boxes <= 8096, after which it matches numpy version).

    Args:
        boxes: np.ndarray
            bounding boxes in cornerform
        scores: np.ndarray
            prediction confidence
        threshold: float
            minimum iou threshold required to retain boxes
    """
    n, _ = boxes.shape

    # compute area
    areas = np.zeros(n, dtype=np.float32)
    for i in numba.prange(n):  # pylint: disable=not-an-iterable
        areas[i] = (boxes[i, 2] - boxes[i, 0] + 1) * (boxes[i, 3] - boxes[i, 1] + 1)

    order = scores.argsort()[::-1]
    keep = np.ones(n, dtype=np.int32)
    for i in range(n):
        if not keep[order[i]]:
            continue

        for j in numba.prange(i + 1, n):  # pylint: disable=not-an-iterable
            if not keep[order[j]]:
                continue

            w = max(0.0, min(boxes[order[i], 2], boxes[order[j], 2]) - max(boxes[order[i], 0], boxes[order[j], 0]) + 1)
            h = max(0.0, min(boxes[order[i], 3], boxes[order[j], 3]) - max(boxes[order[i], 1], boxes[order[j], 1]) + 1)
            intersection = w * h
            iou = intersection / (areas[order[i]] + areas[order[j]] - intersection)
            if iou > threshold:
                keep[order[j]] = 0

    return keep



from typing import Annotated, Literal

import cv2
import numpy as np
import pydantic

from astrapia.data.detection import BaseDetection
from astrapia.geometry import transform


class Face(BaseDetection):
    name: Annotated[Literal["FACE"], pydantic.Field(default="FACE")]

    __target_corners__ = np.array([[0.25, 0.35], [0.75, 0.35], [0.75, 0.85], [0.25, 0.85]], dtype=np.float32)
    __target_points__ = np.array([[0.39, 0.50], [0.61, 0.50], [0.50, 0.65], [0.50, 0.75]], dtype=np.float32)

    @pydantic.field_validator("box", "embedding", mode="before")
    @classmethod
    def validate_1d_ndarray(cls, data: np.ndarray) -> np.ndarray:
        if isinstance(data, str):
            data = cls.decode(data)
        if isinstance(data, list | tuple):
            data = np.array(data, dtype=np.float32)
        if isinstance(data, np.ndarray):
            data = np.float32(data).reshape(-1)
        return data

    @pydantic.field_validator("points", mode="before")
    @classmethod
    def validate_points(cls, data: np.ndarray) -> np.ndarray:
        if isinstance(data, str):
            data = cls.decode(data)
        if isinstance(data, list | tuple):
            data = np.array(data, dtype=np.float32)
        if isinstance(data, np.ndarray):
            data = np.float32(data).reshape(-1, 2)

        if not (isinstance(data, np.ndarray) and data.shape in ((6, 2), (468, 2))):
            raise ValueError(f"{cls.__name__}: points ndarray of shape (6, 2) or (468, 2).")
        return data

    @property
    def n_points(self) -> int:
        return self.points.shape[0]

    @property
    def iod(self) -> float:
        """Inter ocular distance."""
        return ((self.right_eye - self.left_eye) ** 2).sum().item() ** 0.5

    @property
    def iod_approximate(self) -> float:
        """Approximate IOD better for off angle faces."""
        points = self.source[:3]
        return (((points[:, None] - points[None]) ** 2).sum(-1) ** 0.5).max()

    @property
    def eye_center(self) -> np.ndarray:
        """Eye center (x, y)."""
        return (self.right_eye + self.left_eye) / 2

    @property
    def right_eye(self) -> np.ndarray:
        """Right eye center (x, y)."""
        return self.index2point(self.right_eye_index)

    @property
    def right_eye_index(self) -> list[int]:
        """Right eye center indices."""
        return [0] if self.n_points == 6 else [243, 27, 130, 23]

    @property
    def left_eye(self) -> np.ndarray:
        """Left eye center (x, y)."""
        return self.index2point(self.left_eye_index)

    @property
    def left_eye_index(self) -> list[int]:
        """Left eye center indices."""
        return [1] if self.n_points == 6 else [463, 257, 359, 253]

    @property
    def nose_tip(self) -> np.ndarray:
        """Nose tip (x, y)."""
        return self.index2point(self.nose_tip_index)

    @property
    def nose_tip_index(self) -> list[int]:
        """Nose tip indices."""
        return [2] if self.n_points == 6 else [1, 4, 44, 274]

    @property
    def mouth(self) -> np.ndarray:
        """Mouth center (x, y)."""
        return self.index2point(self.mouth_index)

    @property
    def mouth_index(self) -> list[int]:
        """Mouth center indices."""
        return [3] if self.n_points == 6 else [0, 76, 17, 306]

    def index2point(self, index: tuple[int]):
        """Get points from indices."""
        return None if len(index) == 0 else self.points[index].mean(0)

    @property
    def source(self) -> np.ndarray:
        return np.stack((self.right_eye, self.left_eye, self.nose_tip, self.mouth), 0)

    def aligned_face(
        self,
        image: np.ndarray,
        side: int | None = None,
        pad: float = 0.0,
        allow_smaller_side: bool = True,
    ) -> np.ndarray:
        """Align face with landmarks (eyes, nose-tip and mouth)."""
        source = np.stack((self.right_eye, self.left_eye, self.nose_tip, self.mouth), 0)
        reye, leye, *_ = self.__target_points__
        h = w = int(self.iod_approximate / (((reye - leye) ** 2).sum() ** 0.5))
        if side is not None:
            h = w = h if allow_smaller_side and side > h else side

        target = self.__target_points__.copy()[: source.shape[0]]
        if pad != 0.0:
            target += pad / 2
            target /= 1 + pad
        target[:, 0] *= w
        target[:, 1] *= h
        tm = transform.similarity(source, target)
        aligend_image, _ = transform.source2target_converter(image, None, size_hxw=(h, w), tmat=tm)
        self.storage["tmat"] = tm
        return aligend_image

    def face_crop(self, image: np.ndarray, iod_multiplier: float = 4.0) -> np.ndarray:
        """Face crop with eyes at the center."""
        x, y = self.eye_center.tolist()
        side = self.iod_approximate * iod_multiplier
        x1, y1, x2, y2 = (int(xy) for xy in (x - side / 2, y - side / 2, x + side / 2, y + side / 2))
        return image[y1:y2, x1:x2]

    def aligned_corners(self) -> np.array:
        """Aligned corners."""
        # transformation matrix
        source = np.stack((self.right_eye, self.left_eye), 0)
        reye, leye, *_ = self.__target_points__
        scale = int(self.iod / (((reye - leye) ** 2).sum() ** 0.5))
        target = self.__target_points__.copy()[: source.shape[0]]
        tm = transform.similarity(source, target * scale)

        # convert target corners to image space
        target = self.__target_corners__.copy() * scale
        _, source = transform.source2target_converter(None, points=target, tmat=tm, invert=True)
        return source

    def annotate(self, image: np.ndarray, inplace: bool = False, all_points: bool = False) -> np.ndarray:
        """Annotate image."""
        image = super().annotate(image, inplace=inplace)
        # add points
        radius = int(max(2, self.iod_approximate // 64))
        for x, y in self.points if all_points else (self.right_eye, self.left_eye, self.nose_tip):
            cv2.circle(image, (round(x), round(y)), radius, (16, 196, 146), -1)
        return image

    def __repr__(self) -> str:
        is_mesh = self.points.shape[0] == 468
        box = np.round(self.cornerform).astype(int).tolist()
        return f"Face({self.confidence:.4f}, box={box}, iod={self.iod:.4f}, is_mesh={is_mesh})"



import logging
import pathlib
from abc import ABC, abstractmethod
from typing import Annotated, Any

import numpy as np
import pydantic

from astrapia import assets
from astrapia.callbacks.base import BaseCallback
from astrapia.utils import timer


logger = logging.getLogger("Base")


class Extra(pydantic.BaseModel, extra="allow"): ...


class Base(ABC):
    """Base process."""

    class Specs(pydantic.BaseModel, extra="ignore"):
        name: str
        version: Annotated[str, pydantic.StringConstraints(strip_whitespace=True, to_lower=True)]
        clear_storage: Annotated[bool, pydantic.Field(default=True, frozen=True)]
        extra: Annotated[Extra, pydantic.Field(default=Extra())]

        @pydantic.model_validator(mode="before")
        @classmethod
        def extras(cls, data: dict[str, Any]) -> dict[str, Any]:
            for key in list(data.keys()):
                if key not in cls.model_fields:
                    if "extra" not in data:
                        data["extra"] = {}
                    data["extra"][key] = data[key]
            return data

    __callbacks__: tuple[BaseCallback, ...] = ()
    __extra__ = None
    __requests__: tuple[Any, ...] = ()
    __response__: Any = None
    __specs__: pydantic.BaseModel = Specs

    def __init__(self, **kwargs) -> None:
        self.specs = self.__specs__(**self.default_specs(**kwargs))

    @property
    def name(self) -> str:
        return f"{self.specs.name}: ({self.specs.version})"

    @classmethod
    def from_yaml(cls, path_to_assets: pathlib.Path, path_to_yaml: pathlib.Path):
        if not (isinstance(path_to_assets, pathlib.Path) or path_to_assets is None):
            logger.error(f"{cls.__name__}: path_to_assets must be None | pathlib.Path - {path_to_assets}")
            raise TypeError(f"{cls.__name__}: path_to_assets must be None | pathlib.Path - {path_to_assets}")

        if not isinstance(path_to_yaml, pathlib.Path):
            logger.error(f"{cls.__name__}: path_to_yaml must be pathlib.Path - {path_to_yaml}")
            raise TypeError(f"{cls.__name__}: path_to_yaml must be pathlib.Path - {path_to_yaml}")

        specs = assets.load_yaml(path_to_yaml)
        if path_to_assets is not None:
            specs["path_to_assets"] = path_to_assets

        return cls(**specs)

    def validate_request(self, request: Any) -> None:
        if not isinstance(request, self.__requests__):
            logger.error(f"{self.specs.name}: invalid type for request.")
            raise TypeError(f"{self.specs.name}: invalid type for request.")

    def __call__(self, *args) -> Any:
        if len(args) == 0:
            logger.error(f"{self.specs.name}: missing request.")
            raise TypeError(f"{self.specs.name}: missing request.")
        if len(args) == 1:
            with timer.Timer(name=self.name):
                return self.__process__(*args)
        return [self.__process__(request=arg) for arg in args]

    def __process__(self, request: Any) -> Any:
        self.validate_request(request)
        # callbacks - before
        for callback in self.__callbacks__:
            callback.before_process(request)
        # process
        response = self.process(request)
        # callbacks - after
        for callback in self.__callbacks__:
            callback.after_process(response)
        # clear storage
        if self.specs.clear_storage and hasattr(request, "clear_storage"):
            request.clear_storage()
        if self.specs.clear_storage and hasattr(response, "clear_storage"):
            response.clear_storage()
        return response

    @abstractmethod
    def process(self, request: Any) -> Any: ...

    @abstractmethod
    def default_response(self) -> Any: ...

    def default_specs(self, **kwargs) -> dict[str, Any]:
        return kwargs

    @staticmethod
    def sigmoid(x: np.ndarray) -> np.ndarray:
        pos = x > 0
        neg = ~pos
        x[pos] = 1 / (1 + np.exp(-x[pos]))
        x[neg] = np.exp(x[neg]) / (1 + np.exp(x[neg]))
        return x

    @staticmethod
    def softmax(x: np.ndarray) -> np.ndarray:
        xexp = np.exp(x - x.max(-1, keepdims=True))
        return xexp / xexp.sum(-1, keepdims=True)

    def __repr__(self) -> str:
        return f"{self.specs.name}{self.specs.model_dump_json(exclude={'clear_storage'}, indent=2)}"



from typing import Any, Literal

import numpy as np
import pydantic

from astrapia import assets
from astrapia.callbacks.base import BaseCallback
from astrapia.callbacks.to_bchw import ToBCHW
from astrapia.data.face import Face
from astrapia.data.tensor import ImageTensor
from astrapia.engine.base_ml import BaseML
from astrapia.geometry.transform import source2target_converter


class PrepareImages(BaseCallback):
    def __init__(self, specs: pydantic.BaseModel) -> None:
        self.specs = specs

    def before_process(self, request: ImageTensor) -> Any:
        request.storage["images"] = []
        request.storage["tms"] = []
        for detection in request.detections:
            if isinstance(detection, Face):
                image = detection.aligned_face(
                    request.tensor,
                    side=max(self.specs.size),
                    pad=-0.1,
                    allow_smaller_side=False,
                )
                request.storage["images"].append(image)
                request.storage["tms"].append(detection.storage["tmat"])
            else:
                request.storage["images"].append(None)
                request.storage["tms"].append(None)
        return request


class Algorithm(BaseML):
    class Specs(BaseML.Specs):
        name: Literal["MediaPipe-Mesh"]
        version: Literal["mesh"]

    __specs__ = Specs
    __requests__ = (ImageTensor,)
    __response__ = ImageTensor

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.__callbacks__ += (PrepareImages(self.specs), ToBCHW(self.specs))

    @classmethod
    def load_mesh(cls, engine: Literal["coreml", "onnxruntime"]):
        specs = assets.load_yaml(assets.PATH_TO_ASSETS / "configs/mediapipe_mesh.yaml")
        specs["path_to_assets"] = assets.PATH_TO_ASSETS
        specs["engine"] = engine
        return cls(**specs)

    def process(self, request: Any) -> Any:
        for tensor, tm_image2crop, face in zip(
            request.storage["tensors"],
            request.storage["tms"],
            request.detections,
            strict=False,
        ):
            if tensor is None or tm_image2crop is None:
                continue

            # infer model
            output = super().process(tensor)
            output = {key: np.squeeze(value) for key, value in output.items()}
            score, points_on_crop = output[self.__onames__[0]], output[self.__onames__[1]][..., :2]

            # not reliable
            if score < self.specs.threshold:
                continue
            # project points on crop to image
            _, points_on_image = source2target_converter(None, points_on_crop, tmat=tm_image2crop, invert=True)
            face.points = np.float32(points_on_image)

        return request

    def default_response(self, **kwargs) -> Any:
        # Since request is the response!
        raise NotImplementedError()



__all__ = ["MODELS", "PATH_TO_ASSETS"]

import pathlib
from typing import Any

import yaml


PATH_TO_ASSETS = pathlib.Path(__file__).resolve().parent
MODELS = {
    "coreml-long": PATH_TO_ASSETS / "long.mlpackage",
    "coreml-short": PATH_TO_ASSETS / "short.mlpackage",
    "onnxruntime-long": PATH_TO_ASSETS / "long.onnx",
    "onnxruntime-short": PATH_TO_ASSETS / "short.onnx",
}


def load_yaml(path_to_yaml: pathlib.Path) -> dict[str, Any]:
    with open(path_to_yaml) as f:
        data = yaml.safe_load(f)
    return data



import numpy as np

import astrapia
import astrapia.geometry


def test_similarity_transform():
    r"""Test geometry.transform.similarity."""
    tmat = astrapia.geometry.transform.similarity([[4, 6], [6, 6]], [[2, 3], [3, 3]])
    assert np.allclose(tmat, [[0.5, 0, 0], [0, 0.5, 0], [0, 0, 1]])



import abc
from typing import Any


class BaseCallback(abc.ABC):
    @abc.abstractmethod
    def __init__(self) -> None: ...

    def before_process(self, request: Any) -> Any:
        return request

    def after_process(self, response: Any) -> Any:
        return response



import pathlib
import platform

import cv2

import astrapia


PATH = pathlib.Path(__file__).resolve().parent
(PATH / "results").mkdir(parents=True, exist_ok=True)


def test_short():
    r"""Test mediapipe.mesh."""
    if platform.system() == "Darwin":
        astrapia.utils.timer.ENABLE_TIMER = True
        process = astrapia.examples.mediapipe.detector.Algorithm.load_short(add_mesh=True, engine="coreml")
        astrapia.utils.timer.ENABLE_TIMER = False
        for _ in range(4):  # warmup
            process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

        astrapia.utils.timer.ENABLE_TIMER = True
        # short + mesh
        itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
        process(itensor)
        assert len(itensor.detections) == 1
        assert itensor.detections[0].points.shape[0] == 468
        image = itensor.detections[0].aligned_face(itensor.tensor, 128)
        cv2.imwrite(str(PATH / "results/sample_short_mesh_aligned_coreml.webp"), image[..., ::-1])
        astrapia.utils.timer.ENABLE_TIMER = False

    # onnxruntime
    astrapia.utils.timer.ENABLE_TIMER = True
    process = astrapia.examples.mediapipe.detector.Algorithm.load_short(engine="onnxruntime")
    astrapia.utils.timer.ENABLE_TIMER = False
    # just short
    for _ in range(2):  # warmup
        process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

    astrapia.utils.timer.ENABLE_TIMER = True
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 6, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_short_aligned.webp"), image[..., ::-1])

    # short + mesh
    process = astrapia.examples.mediapipe.detector.Algorithm.load_short(add_mesh=True, engine="onnxruntime")
    astrapia.utils.timer.ENABLE_TIMER = False
    for _ in range(2):  # warmup
        process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

    astrapia.utils.timer.ENABLE_TIMER = True
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 468, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_short_mesh_aligned.webp"), image[..., ::-1])
    itensor.save_detections(PATH / "results/sample_short_mesh.yaml")
    astrapia.utils.timer.ENABLE_TIMER = False


def test_long():
    r"""Test mediapipe.mesh."""
    if platform.system() == "Darwin":
        astrapia.utils.timer.ENABLE_TIMER = True
        process = astrapia.examples.mediapipe.detector.Algorithm.load_long(add_mesh=False, engine="coreml")
        astrapia.utils.timer.ENABLE_TIMER = False
        for _ in range(10):  # warmup
            process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

        astrapia.utils.timer.ENABLE_TIMER = True
        # just long
        itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
        process(itensor)
        assert len(itensor.detections) == 1, itensor.detections
        assert itensor.detections[0].points.shape[0] == 6, itensor.detections[0].model_dump_json(indent=2)
        image = itensor.detections[0].aligned_face(itensor.tensor, 128)
        cv2.imwrite(str(PATH / "results/sample_long_aligned_coreml.webp"), image[..., ::-1])

    # onnxruntime
    astrapia.utils.timer.ENABLE_TIMER = True
    process = astrapia.examples.mediapipe.detector.Algorithm.load_long(engine="onnxruntime")
    astrapia.utils.timer.ENABLE_TIMER = False
    # just long
    for _ in range(2):  # warmup
        process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

    astrapia.utils.timer.ENABLE_TIMER = True
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 6, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_long_aligned.webp"), image[..., ::-1])

    # long + mesh
    process = astrapia.examples.mediapipe.detector.Algorithm.load_long(add_mesh=True, engine="onnxruntime")
    astrapia.utils.timer.ENABLE_TIMER = False
    for _ in range(2):  # warmup
        process(astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg"))

    astrapia.utils.timer.ENABLE_TIMER = True
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 468, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_long_mesh_aligned.webp"), image[..., ::-1])
    itensor.save_detections(PATH / "results/sample_long_mesh.yaml")
    astrapia.utils.timer.ENABLE_TIMER = False


def test_short_2x():
    r"""Test mediapipe.mesh."""
    # onnxruntime
    process = astrapia.examples.mediapipe.detector.Algorithm.load_short(add_mesh=True, engine="onnxruntime")
    process.add_extra_size((1024, 1024))

    # short + mesh
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 468, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_short_mesh_aligned_2x.webp"), image[..., ::-1])


def test_long_2x():
    r"""Test mediapipe.mesh."""
    # onnxruntime
    process = astrapia.examples.mediapipe.detector.Algorithm.load_long(add_mesh=True, engine="onnxruntime")
    process.add_extra_size((1024, 1024))

    # long + mesh
    itensor = astrapia.data.ImageTensor(tensor=PATH / "sample/sample.jpeg")
    process(itensor)
    assert len(itensor.detections) == 1, itensor.detections
    assert itensor.detections[0].points.shape[0] == 468, itensor.detections[0].model_dump_json(indent=2)
    image = itensor.detections[0].aligned_face(itensor.tensor, 128)
    cv2.imwrite(str(PATH / "results/sample_long_mesh_aligned_2x.webp"), image[..., ::-1])



import numpy as np

import astrapia


def test_tensor_encoding_decoding():
    r"""Test astrapia.data.Tensor."""
    dtypes_int = ("int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64")
    dtypes_float = ("float16", "float32", "float64")
    for dtype in dtypes_int + dtypes_float:
        for ndim in range(1, 8):
            shape = np.random.randint(0, 10, ndim).tolist()
            if "int" in dtype:
                data = np.random.randint(0, 100, shape).astype(dtype)
            else:
                data = np.random.randn(*shape).astype(dtype)

            assert np.allclose(data, astrapia.data.BaseTensor.decode(astrapia.data.BaseTensor.encode(data)))



import base64
from typing import Annotated, Any

import numpy as np
import pydantic


__MAGIC__: str = "@$tr@46="


class BaseData(pydantic.BaseModel, arbitrary_types_allowed=True, extra="ignore"):
    storage: Annotated[dict[str, Any], pydantic.Field(default={})]

    def clear_storage(self) -> None:
        self.storage = {}

    def model_dump(self, **kwargs) -> dict[str, Any]:
        kwargs["exclude"] = kwargs["exclude"] if "exclude" in kwargs else set()
        kwargs["exclude"].add("storage")
        return super().model_dump(**kwargs)

    def model_dump_json(self, **kwargs) -> str:
        self.clear_storage()
        kwargs["exclude"] = kwargs["exclude"] if "exclude" in kwargs else set()
        kwargs["exclude"].add("storage")
        return super().model_dump_json(**kwargs)

    @staticmethod
    def encode(data: np.ndarray) -> str:
        # 'utf-8' ndarray encoding for json serialization.
        if not isinstance(data, np.ndarray):
            raise TypeError("BaseData.encode: tensor must be >= 1-dimension ndarray.")
        if data.ndim == 0:
            raise ValueError("BaseData.encode: tensor must be >= 1-dimension ndarray.")

        magic: str = f"{data.dtype!s}/{__MAGIC__}"
        if len(data.shape) >= 2:
            magic = f"{'/'.join(str(n) for n in data.shape)}/{magic}"
        return base64.b64encode(magic.encode() + data.tobytes()).decode("utf-8")

    @staticmethod
    def decode(data: str) -> np.ndarray:
        splits = base64.b64decode(data).split(__MAGIC__.encode())
        if len(splits) != 2:
            raise ValueError("BaseData.decode: data does not contain the MAGIC string.")

        *subsplits, dtype = splits[0].decode().strip()[:-1].split("/")
        shape = tuple(map(int, subsplits)) if len(subsplits) else (-1,)
        return np.frombuffer(splits[1], dtype).reshape(*shape)



import logging
import pathlib
from abc import abstractmethod
from typing import Annotated, Any, Literal

import numpy as np
import pydantic

from astrapia.engine.base import Base


logger = logging.getLogger("BaseML")


class BaseML(Base):
    """Base ML process."""

    class Specs(Base.Specs, arbitrary_types_allowed=True, extra="ignore"):
        path_to_assets: pathlib.Path
        path_in_assets: str
        engine: Annotated[Literal["coreml", "onnxruntime"], pydantic.Field(frozen=True)]
        size: tuple[int, ...]
        mean: np.ndarray | None
        stnd: np.ndarray | None
        interpolation: Literal[1, 2, 3]
        threshold: float

        @pydantic.field_validator("mean", "stnd", mode="before")
        @classmethod
        def to_ndarray(cls, data: Any) -> np.ndarray | None:
            if isinstance(data, list | tuple):
                data = np.squeeze(np.array(data, dtype=np.float32))
            return data

        @pydantic.field_serializer("mean", "stnd", when_used="json")
        def serialize_mean_stnd(self, data: np.ndarray) -> str:
            return data.tolist() if isinstance(data, np.ndarray) else data

    __specs__ = Specs
    __inames__: tuple[str, ...] = ()
    __onames__: tuple[str, ...] = ()
    __onnx_providers__: Literal["AUTO", "CPU"] = "AUTO"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._load_model()

    @property
    def name(self) -> str:
        return f"{self.specs.name}: ({self.specs.version}, engine={self.specs.engine})"

    def _load_model(self) -> None:
        path_to_model = (self.specs.path_to_assets / self.specs.path_in_assets).with_suffix(".onnx")
        if self.specs.engine == "coreml":
            path_to_model = path_to_model.with_suffix(".mlmodel")

        if self.specs.engine == "coreml":  # load coreml model
            if not path_to_model.is_file():
                path_to_model = path_to_model.with_suffix(".mlpackage")
                if not path_to_model.is_dir():
                    logger.error(f"{self.__class__.__name__}: file | folder not found {path_to_model}")
                    raise FileNotFoundError(f"{self.__class__.__name__}: file | folder not found {path_to_model}")

            import coremltools as ct

            self.model = ct.models.MLModel(str(path_to_model))
            # get input and output names
            spec = self.model.get_spec()
            self.__inames__ = tuple(x.name for x in spec.description.input)
            self.__onames__ = tuple(x.name for x in spec.description.output)

        elif self.specs.engine == "onnxruntime":  # load onnxruntime model
            if not path_to_model.is_file():
                logger.error(f"{self.__class__.__name__}: file not found {path_to_model}")
                raise FileNotFoundError(f"{self.__class__.__name__}: file not found {path_to_model}")

            import onnxruntime as ort

            options = ort.SessionOptions()
            options.enable_mem_pattern = True
            options.enable_mem_reuse = True
            options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
            providers = ["CPUExecutionProvider"]
            if self.__onnx_providers__ == "AUTO" and "TensorrtExecutionProvider" in ort.get_available_providers():
                providers = ["TensorrtExecutionProvider", *providers]
            if self.__onnx_providers__ == "AUTO" and "CUDAExecutionProvider" in ort.get_available_providers():
                providers = ["CUDAExecutionProvider", *providers]
            if self.__onnx_providers__ == "AUTO" and "CoreMLExecutionProvider" in ort.get_available_providers():
                providers = ["CoreMLExecutionProvider", *providers]
            self.model = ort.InferenceSession(path_to_model, sess_options=options, providers=providers)
            # get input and output names
            self.__inames__ = tuple(x.name for x in self.model.get_inputs())
            self.__onames__ = tuple(x.name for x in self.model.get_outputs())

    def _call_coreml(self, *args) -> dict[str, Any]:
        output = self.model.predict(dict(zip(self.__inames__, args, strict=False)))
        return {name: output[name] for name in self.__onames__}

    def _call_onnxruntime(self, *args) -> dict[str, Any]:
        output = self.model.run(list(self.__onames__), dict(zip(self.__inames__, args, strict=False)))
        return dict(zip(self.__onames__, output, strict=False))

    def process(self, *args) -> Any:
        if self.specs.engine == "coreml":
            output = self._call_coreml(*args)

        elif self.specs.engine == "onnxruntime":
            output = self._call_onnxruntime(*args)

        else:
            output = None

        return output

    @abstractmethod
    def default_response(self) -> Any: ...



from typing import Any

import numpy as np

from astrapia.callbacks.base import BaseCallback


class ToBCHW(BaseCallback):
    """Requires mean & stnd in specs, and expects 'image' | 'images' in request.storage."""

    def __init__(self, specs: Any) -> None:
        self.specs = specs

    def before_process(self, request: Any) -> Any:
        if (
            "image" in request.storage
            and isinstance(request.storage["image"], np.ndarray)
            and request.storage["image"].ndim in (2, 3)
        ):
            request.storage["tensor"] = self.to_bchw(request.storage["image"])

        if (
            "images" in request.storage
            and isinstance(request.storage["images"], list | tuple)
            and all(isinstance(x, np.ndarray) or x is None for x in request.storage["images"])
            and all(x.ndim in (2, 3) for x in request.storage["images"] if isinstance(x, np.ndarray))
        ):
            request.storage["tensors"] = [x if x is None else self.to_bchw(x) for x in request.storage["images"]]

        return request

    def to_bchw(self, image: np.ndarray) -> np.ndarray:
        image = np.float32(image)
        if hasattr(self.specs, "mean") and isinstance(self.specs.mean, np.ndarray):
            image = image - self.specs.mean
        if hasattr(self.specs, "stnd") and isinstance(self.specs.stnd, np.ndarray):
            image = image / self.specs.stnd

        # HWC to BCHW
        image = image[None, None] if image.ndim == 2 else np.transpose(image, (2, 0, 1))[None]
        return image



__all__ = ["BaseData", "BaseDetection", "BaseTensor", "Face", "ImageTensor"]

from astrapia.data.base import BaseData
from astrapia.data.detection import BaseDetection
from astrapia.data.face import Face
from astrapia.data.tensor import BaseTensor, ImageTensor



from functools import lru_cache
from typing import Any, Literal

import numpy as np
import pydantic

from astrapia import assets
from astrapia.callbacks.base import BaseCallback
from astrapia.callbacks.to_bchw import ToBCHW
from astrapia.data.face import Face
from astrapia.data.tensor import ImageTensor
from astrapia.engine.base_ml import BaseML
from astrapia.examples.mediapipe.mesh import Algorithm as AlgoMesh
from astrapia.utils.nms_fns import nms_numba


class PrepareImages(BaseCallback):
    def __init__(self, specs: pydantic.BaseModel) -> None:
        self.specs = specs

    def before_process(self, request: ImageTensor) -> Any:
        request.storage["images"] = []
        request.storage["scales"] = []
        request.storage["sizes"] = []
        for size in [self.specs.size, *self.specs.extra.sizes]:
            image, scale = request.resize_with_pad(size)
            request.storage["images"].append(image)
            request.storage["scales"].append(scale)
            request.storage["sizes"].append(size)
        return request


class AddMesh(BaseCallback):
    def __init__(self, specs: pydantic.BaseModel) -> None:
        self.specs = specs
        self.mesh_model = AlgoMesh.load_mesh(engine=self.specs.engine)

    def after_process(self, request: ImageTensor) -> Any:
        self.mesh_model(request)
        return request


class Algorithm(BaseML):
    class Specs(BaseML.Specs):
        name: Literal["MediaPipe-Detector"]
        version: Literal["short", "long"]

    __specs__ = Specs
    __requests__ = (ImageTensor,)
    __response__ = ImageTensor

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.__callbacks__ += (PrepareImages(self.specs), ToBCHW(self.specs))
        if self.specs.extra.add_mesh:
            self.__callbacks__ += (AddMesh(self.specs),)

    def add_extra_size(self, size: tuple[int, int]) -> None:
        self.specs.extra.sizes.add(size)

    def _compute_anchors(self, size: tuple[int, int]) -> np.ndarray:
        @lru_cache(maxsize=4)
        def _compute_anchors(size: tuple[int, int], strides: tuple[int, ...]):
            h, w = size
            anchors: list[np.ndarray] = []
            for st in np.sort(np.unique(strides)).tolist():
                repeats = (np.array(strides) == st).sum() * 2
                nh = h // st
                nw = w // st
                ys = np.linspace(0.5, nh - 0.5, nh) / nh
                xs = np.linspace(0.5, nw - 0.5, nw) / nw
                yxs = np.stack(np.meshgrid(ys, xs), axis=-1).reshape(-1, 2)
                anchors.append(yxs.repeat(repeats, axis=0))
            return np.concatenate(anchors).astype(np.float32)

        return _compute_anchors(size, self.specs.extra.strides)

    @classmethod
    def load_short(cls, engine: Literal["coreml", "onnxruntime"], **kwargs):
        specs = assets.load_yaml(assets.PATH_TO_ASSETS / "configs/mediapipe_short.yaml")
        specs["path_to_assets"] = assets.PATH_TO_ASSETS
        specs["engine"] = engine
        return cls(**specs, **kwargs)

    @classmethod
    def load_long(cls, engine: Literal["coreml", "onnxruntime"], **kwargs):
        specs = assets.load_yaml(assets.PATH_TO_ASSETS / "configs/mediapipe_long.yaml")
        specs["path_to_assets"] = assets.PATH_TO_ASSETS
        specs["engine"] = engine
        return cls(**specs, **kwargs)

    def process(self, request: Any) -> Any:
        all_scores, all_boxes, all_points = None, None, None
        for tensor, scale, size in zip(
            request.storage["tensors"],
            request.storage["scales"],
            request.storage["sizes"],
            strict=False,
        ):
            # infer model
            output = super().process(tensor)
            output = {key: np.squeeze(value) for key, value in output.items()}
            anchors = self._compute_anchors(size)

            # decode points
            scores, regression = output["scores"], output["regression"]
            num_points = regression.shape[-1] // 2
            regression = regression.reshape(-1, num_points, 2) / max(size)
            regression += anchors[:, None, :]
            regression[:, 1] -= anchors

            # convert x_center, y_center, w, h to xmin, ymin, xmax, ymax
            # split into boxes and points
            boxes = np.zeros_like(regression[:, :2])
            boxes[:, 0] = regression[:, 0] - regression[:, 1] / 2
            boxes[:, 1] = regression[:, 0] + regression[:, 1] / 2
            points = regression[:, 2:]
            boxes = np.clip(boxes * max(size), 0, None)
            boxes = boxes.reshape(-1, 4)
            points = np.clip(points * max(size), 0, None)
            # thresholding
            threshold = self.specs.threshold if size == self.specs.size else self.specs.extra.threshold_extra_sizes
            valid = scores >= threshold
            scores, boxes, points = (x[valid] for x in (scores, boxes, points))

            # scale to image size
            boxes /= scale
            points /= scale

            # aggregate all
            all_scores = scores.copy() if all_scores is None else np.concatenate((all_scores, scores))
            all_boxes = boxes.copy() if all_boxes is None else np.concatenate((all_boxes, boxes))
            all_points = points.copy() if all_points is None else np.concatenate((all_points, points))

        # highest score first
        sort = np.flip(np.argsort(all_scores))
        all_scores, all_boxes, all_points = (x[sort] for x in (all_scores, all_boxes, all_points))
        # remove invalid area
        if self.specs.extra.min_area is not None:
            valid = (((all_boxes[..., 2:] - all_boxes[..., :2]) ** 2).sum(-1) ** 0.5) >= self.specs.extra.min_area
            all_scores, all_boxes, all_points = (x[valid] for x in (all_scores, all_boxes, all_points))
        # nms
        valid = nms_numba(all_boxes, all_scores, self.specs.extra.threshold_iou) == 1
        all_scores, all_boxes, all_points = (x[valid] for x in (all_scores, all_boxes, all_points))

        # add all detections
        for confidence, box, points in zip(all_scores.tolist(), all_boxes, all_points, strict=False):
            request.detections.append(Face(label="FACE", confidence=confidence, box=box, points=points))
        return request

    def default_response(self, **kwargs) -> Any:
        # Since request is the response!
        raise NotImplementedError()

    def default_specs(self, **kwargs) -> dict[str, Any]:
        kwargs.setdefault("version", "short")
        kwargs.setdefault("sizes", set())
        kwargs.setdefault("add_mesh", False)
        kwargs["strides"] = (8, 16, 16, 16) if kwargs["version"] == "short" else (16, 32, 32, 32)
        kwargs.setdefault("threshold_iou", 0.3)
        kwargs.setdefault("threshold_extra_sizes", 0.8)
        kwargs.setdefault("min_area", 12)
        return kwargs



__all__ = ["similarity", "source2target_converter"]

import cv2
import numpy as np


def similarity(source: np.ndarray, target: np.ndarray) -> np.ndarray:
    """Compute similarty transform to convert source to target."""
    # This function is ported from Scikit-Image.
    # Copyright: 2009-2022 the scikit-image team
    # License: BSD-3-Clause (https://scikit-image.org/docs/stable/license.html)
    # https://github.com/scikit-image/scikit-image/blob/v0.23.2/skimage/transform/_geometric.py
    source, target = map(np.float64, (source, target))
    dim = source.shape[1]
    s_mu, t_mu = source.mean(axis=0), target.mean(axis=0)
    A = (target - t_mu).T @ (source - s_mu) / source.shape[0]
    valid = np.ones((source.shape[1],), dtype=np.float64)
    if np.linalg.det(A) < 0:
        valid[dim - 1] = -1

    T = np.eye(dim + 1, dtype=np.float64)
    U, S, V = np.linalg.svd(A)
    rank = np.linalg.matrix_rank(A)
    if rank == 0:
        return np.nan * T

    if rank == dim - 1:
        if np.linalg.det(U) * np.linalg.det(V) > 0:
            T[:dim, :dim] = U @ V

        else:
            save = valid[dim - 1]
            valid[dim - 1] = -1
            T[:dim, :dim] = U @ np.diag(valid) @ V
            valid[dim - 1] = save

    else:
        T[:dim, :dim] = U @ np.diag(valid) @ V

    scale = 1.0 / (source - s_mu).var(axis=0).sum() * (S @ valid)
    T[:dim, dim] = t_mu - scale * (T[:dim, :dim] @ s_mu.T)
    T[:dim, :dim] *= scale
    return T


def source2target_converter(
    image: np.ndarray | None,
    points: np.ndarray | None,
    size_hxw: tuple[int, int] | None = None,
    source: np.ndarray | None = None,
    target: np.ndarray | None = None,
    tmat: np.ndarray | None = None,
    invert: bool = False,
) -> tuple[np.ndarray | None, np.ndarray | None]:
    if not ((isinstance(source, np.ndarray) and isinstance(target, np.ndarray)) or isinstance(tmat, np.ndarray)):
        raise ValueError("source2target_converter: Neither tmat nor (source and target) are valid.")
    if tmat is None:  # only compute when tmat is not available
        n = min(source.shape[0], target.shape[0])
        tmat = similarity(source[:n], target[:n])
    if invert:  # projecting target (image | points) to source
        tmat = np.linalg.inv(tmat)

    if image is not None:  # warp image
        if size_hxw is None:
            raise ValueError("source2target_converter: size_hxw is required to transform image.")
        image = cv2.warpAffine(image, tmat[:2], size_hxw[::-1])

    if points is not None:  # project points
        points = np.concatenate((points, np.ones(points.shape[0])[:, None]), -1)
        points = (tmat @ points.T).T
        points = points[:, :2] / points[:, [2]]
    return image, points



__all__ = ["Timer", "timer"]

import logging
import time
from collections.abc import Callable
from functools import wraps
from typing import Any


ENABLE_TIMER: bool = False
logger = logging.getLogger("Timer")


class Timer:
    def __init__(self, name: str):
        self.name = name

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args, **kwargs):
        total = time.time() - self.start
        if ENABLE_TIMER:
            logger.info(f"{self.name} took {total:2.4f} seconds.")
            print(f"{self.name} took {total:2.4f} seconds.")  # noqa: T201


def timer(fn: Callable, name: str) -> Callable:
    @wraps(fn)
    def fn_wrap(*args, **kwargs) -> Any:
        with Timer(name=name):
            result = fn(*args, **kwargs)
        return result

    return fn_wrap



from typing import Annotated

import cv2
import numpy as np
import pydantic

from astrapia.data.base import BaseData


class BaseDetection(BaseData, arbitrary_types_allowed=True):
    label: str
    confidence: float
    box: np.ndarray
    embedding: Annotated[np.ndarray | None, pydantic.Field(default=None)]
    points: Annotated[np.ndarray | None, pydantic.Field(default=None)]

    @pydantic.field_validator("box", "embedding", "points", mode="before")
    @classmethod
    def validate_1d_ndarray(cls, data: np.ndarray) -> np.ndarray:
        if isinstance(data, str):
            data = cls.decode(data)
        if isinstance(data, list | tuple):
            data = np.array(data, dtype=np.float32)
        if isinstance(data, np.ndarray):
            data = np.float32(data).reshape(-1)
        return data

    @property
    def centerform(self) -> np.ndarray:
        return self.cornerform_to_centerform(self.cornerform)

    @property
    def cornerform(self) -> np.ndarray:
        return self.box.copy()

    @property
    def corners(self) -> tuple[tuple[float, float], ...]:
        x1, y1, x2, y2 = self.cornerform.tolist()
        return (
            (x1, y1),
            (x2, y1),
            (x2, y2),
            (x1, y2),
        )

    def crop(self, image: np.ndarray, scale: float = 1.0) -> np.ndarray:
        """Image crop using detected box."""
        x1, y1, x2, y2 = np.round(self.crop_xyxy(scale=scale)).astype(int).tolist()
        return image[y1:y2, x1:x2]

    def crop_xyxy(self, scale: float = 1.0) -> np.ndarray:
        """Detected box."""
        x, y, w, h = self.centerform.tolist()
        w *= scale
        h *= scale
        return self.centerform_to_cornerform((x, y, w, h))

    def annotate(self, image: np.ndarray, inplace: bool = False) -> np.ndarray:
        """Annotate image."""
        image = image if inplace else image.copy()
        x1, y1, x2, y2 = np.round(self.cornerform).astype(int).tolist()
        cv2.rectangle(image, (x1, y1), (x2, y2), color=(236, 46, 36), thickness=2)
        return image

    @staticmethod
    def cornerform_to_centerform(box: np.ndarray) -> np.ndarray:
        x1, y1, x2, y2 = np.float32(box)
        return np.float32([(x1 + x2) / 2, (y1 + y2) / 2, x2 - x1, y2 - y1])

    @staticmethod
    def centerform_to_cornerform(box: np.ndarray) -> np.ndarray:
        x, y, w, h = np.float32(box)
        return np.float32([x - w / 2, y - h / 2, x + w / 2, y + h / 2])

    @pydantic.field_serializer("box", "embedding", "points", when_used="json")
    def serialize_ndarray(self, data: np.ndarray | None) -> str | None:
        return data if data is None else self.encode(data)

    def __repr__(self) -> str:
        box = np.round(self.cornerform).astype(int).tolist()
        return f"Detection({self.confidence:.4f}, box={box})"



import os
import pathlib
from typing import Annotated

import cv2
import numpy as np
import pydantic
import yaml
from PIL import Image as ImPIL

from astrapia.data.base import BaseData
from astrapia.data.detection import BaseDetection
from astrapia.data.face import Face


class BaseTensor(BaseData, arbitrary_types_allowed=True):
    """Tensor: Numpy ndarray with json serialization."""

    tensor: np.ndarray

    @pydantic.field_validator("tensor", mode="before")
    @classmethod
    def validate_tensor(cls, data: np.ndarray | str) -> np.ndarray:
        if isinstance(data, str):
            data = BaseData.decode(data)
        return data

    @pydantic.field_serializer("tensor", when_used="json")
    def serialize_tensor(self, data: np.ndarray) -> str:
        return BaseData.encode(data)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.tensor.shape


class ImageTensor(BaseTensor):
    detections: Annotated[list[BaseDetection | Face], pydantic.Field(default=[])]

    @pydantic.field_validator("tensor", mode="before")
    @classmethod
    def validate_tensor(cls, data: np.ndarray | pathlib.Path | str) -> np.ndarray:
        if isinstance(data, pathlib.Path):
            data = cv2.cvtColor(cv2.imread(str(data), cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)
        if isinstance(data, str):
            data = cv2.cvtColor(cv2.imread(data, 1), 4) if os.path.isfile(data) else BaseTensor.decode(data)

        if not (isinstance(data, np.ndarray) and (data.ndim == 2 or (data.ndim == 3 and data.shape[-1] in (1, 3)))):
            raise ValueError(f"{cls.__name__}: tensor must be 2/3 dimensional ndarray.")
        return data

    @pydantic.field_validator("detections", mode="before")
    @classmethod
    def validate_detections(cls, data: list[BaseDetection | Face]) -> np.ndarray:
        if isinstance(data, list | tuple):
            data = [Face(**x) if isinstance(x, dict) and x["label"] == "FACE" else x for x in data]
            data = [BaseDetection(**x) if isinstance(x, dict) and x["label"] != "FACE" else x for x in data]
        return data

    @property
    def pil_image(self) -> ImPIL.Image:
        return ImPIL.fromarray(self.tensor)

    @property
    def height(self) -> int:
        return self.shape[0]

    @property
    def width(self) -> int:
        return self.shape[1]

    def to_gray(self, inplace: bool = False) -> np.ndarray:
        gray = cv2.cvtColor(self.tensor, cv2.COLOR_RGB2GRAY)
        if inplace:
            self.tensor = gray
        return gray

    def to_bchw(self, gray: bool = False) -> np.ndarray:
        return (self.to_gray()[None, None] if gray else np.transpose(self.tensor.copy(), (2, 0, 1)))[None]

    def annotate(self, **kwargs) -> np.ndarray:
        """Annotate image."""
        image = self.tensor.copy()
        for detection in self.detections:
            detection.annotate(image, inplace=True, **kwargs)
        return image

    def resize(self, size: tuple[int, int], interpolation: int = 3) -> np.ndarray:
        # resize
        #   size: (height x width)
        #   interpolation: 1 is INTER_LINEAR, 2 is INTER_CUBIC, 3 is INTER_AREA
        return cv2.resize(self.tensor, size[::-1], interpolation=interpolation)

    def resize_with_pad(self, size: tuple[int, int], interpolation: int = 3) -> tuple[np.ndarray, float]:
        # resize and keep aspect ratio
        #   size: (height x width)
        #   interpolation: 1 is INTER_LINEAR, 2 is INTER_CUBIC, 3 is INTER_AREA
        scale = size[0] / self.shape[0]
        if any(int(tsz * scale) > sz for tsz, sz in zip(self.shape, size, strict=False)):
            scale = size[1] / self.shape[1]
        size_new = tuple(map(int, map(round, (self.shape[0] * scale, self.shape[1] * scale))))

        tensor = cv2.resize(self.tensor, size_new[::-1], interpolation=interpolation)
        canvas = np.zeros(list(size) + ([self.shape[-1]] if len(self.shape) == 3 else []), tensor.dtype)
        canvas[: size_new[0], : size_new[1]] = tensor
        return canvas, scale

    def load_detections(self, path: pathlib.Path) -> None:
        with open(path) as txt:
            detections = yaml.safe_load(txt.read())["detections"]

        self.detections += ImageTensor.validate_detections(detections)

    def save_detections(self, path: pathlib.Path) -> None:
        if not isinstance(path, pathlib.Path):
            raise TypeError(f"{self.__class__.__name__}.save_detections: path must be pathlib.Path object.")
        if path.suffix != ".yaml":
            raise TypeError(f"{self.__class__.__name__}.save_detections: path must have '.yaml' as suffix.")

        for detection in self.detections:
            detection.clear_storage()
        with open(path, "w") as txt:
            txt.write(self.model_dump_json(exclude={"storage", "tensor"}, indent=2))
